/*********************************************************************
 *
 * Authors: Vincenzo Ciaschini - Vincenzo.Ciaschini@cnaf.infn.it 
 *
 * Copyright (c) 2002, 2003 INFN-CNAF on behalf of the EU DataGrid.
 * For license conditions see LICENSE file or
 * http://www.edg.org/license.html
 *
 * Parts of this code may be based upon or even include verbatim pieces,
 * originally written by other people, in which case the original header
 * follows.
 *
 *********************************************************************/

/* $Revision: 1.1 $ */
/* $Date: 2003/05/13 06:58:01 $ */

#ifndef _VOMS_API_H
#define _VOMS_API_H

#include <openssl/x509.h>

/*! \brief User's characteristics: can be repeated. 
*/
struct data {
  char *group; /*!< user's group */
  char *role;  /*!< user's role */
  char *cap;   /*!< user's capability */
};

/*! \brief The type of data returned.
 */

#define  TYPE_NODATA 0  /*!< no data */
#define  TYPE_STD    1  /*!< group, role, capability triplet */
#define  TYPE_CUSTOM 2  /*!< result of an S command */


struct voms {
  int siglen;         /*!< The length of the VOMS server signature */
  char *signature;   /*!< The VOMS server signature */
  char *user;        /*!< The user's DN, as from his certificate */
  char *userca;      /*!< The CA which signed the user's certificate */
  char *server;      /*!< The VOMS server DN, as from its certificate */
  char *serverca;    /*!< The CA which signed the VOMS certificate */
  char *voname;      /*!< The name of the VO to which the VOMS belongs */
  char *uri;         /*!< The URI of the VOMS server */
  char *date1;       /*!< Beginning of validity of the user info */
  char *date2;       /*!< End of validity of the user info */
  int   type;     /*!< The type of data returned */
  struct data **std;   /*!< User's characteristics */
  char *custom;      /*!< The data returned by an S command */
  int datalen;
  char *reserved;
};


#define RECURSE_CHAIN 0
#define RECURSE_NONE  1

/*! \brief Error codes.
*/

#define VERR_NONE      0
#define VERR_NOSOCKET  1  /*!< Socket problem*/
#define VERR_NOIDENT   2  /*!< Cannot identify itself (certificate problem) */
#define VERR_COMM      3  /*!< Server problem */
#define VERR_PARAM     4  /*!< Wrong parameters*/
#define VERR_NOEXT     5  /*!< VOMS extension missing */
#define VERR_NOINIT    6  /*!< Initialization error */
#define VERR_TIME      7  /*!< Error in time checking */
#define VERR_IDCHECK   8  /*!< User data in extension different from the real
			       ones */
#define VERR_EXTRAINFO 9  /*!< VO name and URI missing */
#define VERR_FORMAT    10 /*!< Wrong data format */
#define VERR_NODATA    11 /*!< Empty extension */
#define VERR_PARSE     12 /*!< Parse error */
#define VERR_DIR       13 /*!< Directory error */
#define VERR_SIGN      14 /*!< Signature error */
#define VERR_SERVER    15 /*!< Unidentifiable VOMS server */
#define VERR_MEM       16 /*!< Memory problems */
#define VERR_VERIFY    17 /*!< Generic verification error*/
#define VERR_TYPE      18 /*!< Returned data of unknown type */



struct vomsdata {
  char *cdir;
  char *vdir;
  struct voms **data; /*!< User's info, as in the certificate extension.
                           It may contain data gathered from more than one
			   VOMS server, */
  char *workvo;     /*!< The value of the -vo option of the voms-proxy-init
		         command */
  char *extra_data; /*!< The data specified by the user with the --include
		         switch. 
                         Note that this field doesn't contain the result of a
			 request to the VOMS server, but instead data specified
			 by the user. 
                         The reason for the introduction of this extension is
			 to let a user include important data into his proxy
			 certificate, like, for example, a kerberos ticket 
                      */
  int volen;
  int extralen;
};

extern struct vomsdata *VOMS_Init(char *voms, char *cert); /*< Initializes a vomsdata structure for use by the other functions.
						       N.B: This is the ONLY way to correctly initialize a vomsdata structure.
						       /param voms The directory which contains the certificates of the VOMS servers
						       /param cert The directory which contains the CA certificates

						       If voms_dir is empty, the value of the environment variable
						       X509_VOMS_DIR is taken

						       If cert_dir is empty, the value of the environment variable
						       X509_CERT_DIR is taken

						       /return NULL for failure, or a pointer to a properly initialized structure. */

extern void VOMS_Destroy(struct vomsdata *vd); /*< Destroys a proper vomsdata structure
						 /param vd The structure to deallocate. */
extern int VOMS_Contact(char *hostname, int port, char *servsubject,
			char *command, struct vomsdata *vd, int *error);  /*!< Contacts a VOMS server to get a certificate

				    It is the equivalent of the voms_proxy_init command, but 
                                    without the --include functionality.
                                    \param hostname FQDN of the VOMS server
                                    \param port the port on which the VOMS server is listening
                                    \param servsubject the subject of the server's certificate
                                    \param command Command
				    \param vd RETURN PARAMETER: contains the data returned by the connection
				    \param error RETURN PARAMETER: Qualifies the error message
                                    \return failure (0) or success (<>0)
				*/
extern int VOMS_Retrieve(X509 *cert, STACK_OF(X509) *chain, int how,
			 struct vomsdata *vd, int *error);  /*!< Extracts the VOMS extension from an X.509 certificate. 

                                                     The function doesn't check the validity of the certificates, 
                                                     but it does check the content of the user data.
                                                     \param cert The certificate with the VOMS extensions
                                                     \param chain The chain of the validation certificates 
                                                           (only the intermediate ones)
                                                     \param how Recursion type
						     \param vd RETURN PARAMETER: contains the data returned by the connection
						     \param error RETURN PARAMETER: Qualifies the error message
						     \return failure (0) or success (<>0)
                                                   */  
extern int VOMS_Import(char *buffer, int buflen, struct vomsdata *vd, int *error); /*!< Converts data from the format used for inclusion 
                                  into a certificate to the internal format

                                     The function does verify the data.
                                     \param buffer contains the data to be converted 
				     \param buflen contains the length of buffer
				     \param vd RETURN PARAMETER: contains the data returned by the connection
				     \param error RETURN PARAMETER: Qualifies the error message
				     \return failure (0) or success (<>0)
                             */

int getMajorVersionNumber(void);
int getMinorVersionNumber(void);
int getPatchVersionNumber(void);

#endif
