#include "std.h"

#include <iostream>
#include <fstream>
#include <string>
#include <vector>
#include <stdio.h>

#include "Preferences.h"
#include "CertInfo.h"
#include "DateTime.h"
#include "MdsQuery.h"
#include "Xrsl.h"
#include "Giis.h"
#include "Target.h"

#define __UI_LIBRARY__
#include "ui_uploader.h"

#include "ngui.h"


extern "C"
int ngsubxx (const vector <string> & xrslfiles,
	     const vector <string> & xrslstrings,
	     const vector <string> & clusterselect,
	     const vector <string> & clusterreject,
	     const vector <string> & giisurls,
	     const string & joblist,
	     const bool dryrun,
	     const bool dumpxrsl,
	     int timeout,
	     int debug,
	     const bool anonymous) {

  if (xrslfiles.empty() && xrslstrings.empty()) {
    cerr << "ngsub: no xrsl input specified" << endl;
    cerr << "Use \"ngsub -help\" for help" << endl;
    return 1;
  }

  vector <Xrsl> xrsllist;
  vector <Giis> giislist;
  vector <Cluster> clusterlist;

  int error = 0;
  ActivateGlobus();

  // setting defaults

  if (timeout == UNDEFINED) timeout = iGetEnv ("NGTIMEOUT");
  if (timeout == UNDEFINED) timeout = iGetDef ("NGTIMEOUT");
  if (timeout == UNDEFINED) timeout = DEFAULT_TIMEOUT;

  if (debug == UNDEFINED) debug = iGetEnv ("NGDEBUG");
  if (debug == UNDEFINED) debug = iGetDef ("NGDEBUG");
  if (debug == UNDEFINED) debug = 0;

  // get user info

  CertInfo user;
  if (!user) {
    error = 1;
    goto errorexit;
  }

  if (debug) {
    cout << "User subject name: " << user.GetSN() << endl;
    cout << "Remaining proxy lifetime: "
	 << Period (user.TimeLeft() / 60) << endl;
  }

  // parse xrsl

  for (vector <string>::const_iterator vsi = xrslfiles.begin();
       vsi != xrslfiles.end(); vsi++) {
    Xrsl xrsl (*vsi, 0);
    if (!xrsl) {
      cerr << "The XRSL file \"" << *vsi << "\" could not be parsed" << endl;
      error = 1;
      goto errorexit;
    }
    bool done = false;
    for (int i = 0; !done; i++) {
      Xrsl axrsl = xrsl[i];
      if (!axrsl)
	done = true;
      else
	xrsllist.push_back (axrsl);
    }
  }

  for (vector <string>::const_iterator vsi = xrslstrings.begin();
       vsi != xrslstrings.end(); vsi++) {
    Xrsl xrsl (*vsi);
    if (!xrsl) {
      cerr << "The XRSL string \"" << *vsi << "\" could not be parsed" << endl;
      error = 1;
      goto errorexit;
    }
    bool done = false;
    for (int i = 0; !done; i++) {
      Xrsl axrsl = xrsl[i];
      if (!axrsl)
	done = true;
      else
	xrsllist.push_back (axrsl);
    }
  }

  for (vector <Xrsl>::iterator vxi = xrsllist.begin();
       vxi != xrsllist.end(); vxi++) {

    if (debug > 1) {
      cout << "Parsed input XRSL:" << endl;
      vxi->Print();
    }

    vxi->Eval();

    if (vxi->Check()) {
      error = 1;
      goto errorexit;
    }
  }

  // find available clusters

  for (vector <string>::const_iterator vsi = clusterselect.begin();
       vsi != clusterselect.end(); vsi++) {
    bool found = false;
    for (vector <Cluster>::iterator cli = clusterlist.begin();
	 !found && cli != clusterlist.end(); cli++)
      if (*vsi == cli->GetName()) found = true;
    if (!found) clusterlist.push_back (*vsi);
  }

  if (clusterlist.empty()) {
    if (GetGiises (giisurls, giislist)) {
      error = 1;
      goto errorexit;
    }

    clusterlist = FindClusters (giislist, user.GetSN(), anonymous,
				timeout, debug);
    if (clusterlist.empty()) {
      cerr << "ngsub: could not retrieve cluster list from giis" << endl;
      error = 1;
      goto errorexit;
    }
  }

  for (vector <string>::const_iterator vsi = clusterreject.begin();
       vsi != clusterreject.end(); vsi++) {
    for (vector <Cluster>::iterator cli = clusterlist.begin();
	 cli != clusterlist.end(); cli++)
      if (cli->GetName() == *vsi) {
	if (debug) cout << "Rejecting cluster: " << *vsi << endl;
	clusterlist.erase (cli);
	break;
      }
  }

  FindClusterInfo (clusterlist, user.GetSN(), anonymous, timeout, debug);

  // do brokering

  srand (time (NULL));

  for (vector <Xrsl>::iterator vxi = xrsllist.begin();
       vxi != xrsllist.end(); vxi++) {

    vector <Target> targetlist;

    for (vector <Cluster>::iterator cli = clusterlist.begin();
	 cli != clusterlist.end(); cli++) {
      for (vector <Queue>::iterator qli = cli->queues.begin();
	   qli != cli->queues.end(); qli++) {

	Target target (&*cli, &*qli);

	if (debug) {
	  cout << "Cluster: " << target.cluster->GetLongName() << endl;
	  cout << "Queue: " << target.queue->GetName() << endl;
	}

	if (target.queue->GetStatus() != "active") {
	  if (debug) cout << "Queue rejected due to status: "
			  << target.queue->GetStatus() << endl;
	  continue;
	}

	if (target.queue->GetUserFreeCpus() == UNDEFINED) {
	  if (debug) cout << "Queue rejected because user not authorized"
			  << endl;
	  continue;
	}

	if (target.queue->GetMaxQueuable() != UNDEFINED &&
	    target.queue->GetQueued() >= target.queue->GetMaxQueuable()) {
	  if (debug) cout << "Queue rejected because the maximum number "
			  << "of queued jobs is reached" << endl;
	  continue;
	}

	string failattr;
	if (vxi->Test (target, failattr)) {
	  error = 1;
	  goto errorexit;
	}

	if (!target) {
	  if (debug) cout << "Queue rejected because it does not match "
			  << "the XRSL specification (" << failattr << ")"
			  << endl;
	  continue;
	}

	if (target.CalculateNeededFileSizes()) {
	  error = 1;
	  goto errorexit;
	}

	if (target.GetRemoteSize() == UNDEFINED) {
	  if (debug) cout << "Queue rejected because no valid location was "
			  << "found for the inputfiles" << endl;
	  continue;
	}

	if (debug > 1) cout << "Total size of remote files: "
			    << target.GetRemoteSize() << " bytes" << endl;
	if (debug > 1) cout << "Total size of local files: "
			    << target.GetLocalSize() << " bytes" << endl;

	if (target.cluster->GetCacheFree() == UNDEFINED) {
	  if (debug > 1) cout << "Total size of files: "
			      << target.GetNeededTotalSize()
			      << " bytes" << endl;

	  if (target.GetNeededTotalSize() > target.queue->GetUserDiskSpace()) {
	    if (debug) cout << "Queue rejected due to insufficient disk space"
			    << endl;
	    continue;
	  }
	}
	else {
	  if (debug > 1) cout << "Total size of files to be cached: "
			      << target.GetNeededCacheSize()
			      << " bytes" << endl;
	  if (debug > 1) cout << "Total size of files in session directory: "
			      << target.GetNeededSessDirSize()
			      << " bytes" << endl;

	  if (target.GetNeededCacheSize() > target.cluster->GetCacheFree()) {
	    if (debug) cout << "Queue rejected due to insufficient cache space"
			    << endl;
	    continue;
	  }
	  if (target.GetNeededSessDirSize() >
	      target.queue->GetUserDiskSpace()) {
	    if (debug) cout << "Queue rejected due to insufficient disk space"
			    << endl;
	    continue;
	  }
	}

	targetlist.push_back (target);
	if (debug) cout << "Queue accepted as possible submission target"
			<< endl;

      }
    }

    bool submitted = false;
    while (!submitted) {

      int freecpus = 0;
      long long int minremotesize = UNDEFINED;
      long long int minlocalsize = UNDEFINED;

      for (vector <Target>::iterator vti = targetlist.begin();
	   vti != targetlist.end(); vti++) {

	if (vti->queue->GetMaxRunning() != UNDEFINED &&
	    vti->queue->GetRunning() >= vti->queue->GetMaxRunning()) continue;

	int time;
	if (vti->xrsl.GetCpuTime (&time)) {
	  error = 1;
	  goto errorexit;
	}
	if (time == UNDEFINED) time = vti->queue->GetDefaultCpuTime();
	int userfreecpus = vti->queue->GetUserFreeCpus (time);

	int count;
	if (vti->xrsl.GetCount (&count)) {
	  error = 1;
	  goto errorexit;
	}
	if (count == UNDEFINED) count = 1;

	if (userfreecpus < count) continue;

	if (minremotesize == UNDEFINED ||
	    vti->GetRemoteSize() < minremotesize) {
	  minremotesize = vti->GetRemoteSize();
	  minlocalsize = vti->GetLocalSize();
	  freecpus = userfreecpus;
	}
	else if (vti->GetRemoteSize() == minremotesize) {
	  if (vti->GetLocalSize() < minlocalsize) {
	    minlocalsize = vti->GetLocalSize();
	    freecpus = userfreecpus;
	  }
	  else
	    freecpus += userfreecpus;
	}
      }

      vector <Target>::iterator target = targetlist.end();

      if (freecpus > 0) {
	float rn = rand() / (RAND_MAX + 1.0) * freecpus;
	int counter = 0;
	bool found = false;

	for (vector <Target>::iterator vti = targetlist.begin();
	     !found && (vti != targetlist.end()); vti++) {

	  if (vti->queue->GetMaxRunning() != UNDEFINED &&
	      vti->queue->GetRunning() >= vti->queue->GetMaxRunning())
	    continue;

	  int time;
	  if (vti->xrsl.GetCpuTime (&time)) {
	    error = 1;
	    goto errorexit;
	  }
	  if (time == UNDEFINED) time = vti->queue->GetDefaultCpuTime();
	  int userfreecpus = vti->queue->GetUserFreeCpus (time);

	  int count;
	  if (vti->xrsl.GetCount (&count)) {
	    error = 1;
	    goto errorexit;
	  }
	  if (count == UNDEFINED) count = 1;

	  if (userfreecpus < count) continue;

	  if (vti->GetRemoteSize() == minremotesize &&
	      vti->GetLocalSize() == minlocalsize) {
	    counter += userfreecpus;
	    if (rn < counter) {
	      target = vti;
	      found = true;
	    }
	  }
	}
      }

      else {
	for (vector <Target>::iterator vti = targetlist.begin();
	     vti != targetlist.end(); vti++) {

	  int count;
	  if (vti->xrsl.GetCount (&count)) {
	    error = 1;
	    goto errorexit;
	  }
	  if (count == UNDEFINED) count = 1;

	  if (vti->GetTotalCpus() < count) continue;

	  if (target == targetlist.end() ||
	      (float) vti->GetQueued() / vti->GetTotalCpus() <
	      (float) target->GetQueued() / target->GetTotalCpus())
	    target = vti;
	}
      }

      if (target == targetlist.end()) {
	cerr << "No cluster found" << endl;
	error = 1;
	goto errorexit;
      }

      if (debug) {
	cout << target->cluster->GetLongName() << " selected" << endl;
	cout << "queue " << target->queue->GetName() << " selected" << endl;
      }


      target->xrsl.AddSimpleRelation ("queue", target->queue->GetName());
      target->xrsl.AddSimpleRelation ("action", "request");
      target->xrsl.AddSimpleRelation ("savestate", "yes");
      target->xrsl.AddSimpleRelation ("clientxrsl", vxi->str());

      if (dryrun) target->xrsl.AddSimpleRelation ("dryrun", "dryrun");

      if ((target->xrsl.FixJoin()) ||
	  (target->xrsl.FixInOut ("executable", "inputfiles")) ||
	  (target->xrsl.FixInOut ("stdin", "inputfiles")) ||
	  (target->xrsl.FixInOut ("stdout", "outputfiles")) ||
	  (target->xrsl.FixInOut ("stderr", "outputfiles")) ||
	  (target->xrsl.FixInOut ("gmlog", "outputfiles")) ||
	  (target->xrsl.FixExec()) ||
	  (target->xrsl.FixStdin()) ||
	  (target->xrsl.FixTime ("starttime")) ||
	  (target->xrsl.FixPeriod ("cputime")) ||
	  (target->xrsl.FixPeriod ("lifetime"))) {
	error = 1;
	goto errorexit;
      }

      char hostname[1024];
      globus_libc_gethostname (hostname, 1024);
      target->xrsl.AddSimpleRelation ("hostname", hostname);

      string jobname;
      if (target->xrsl.GetJobName (jobname)) {
	error = 1;
	goto errorexit;
      }

      vector <string> filenames;
      if (target->xrsl.PrepareUpload (filenames)) {
	error = 1;
	goto errorexit;
      }

      if ((target->xrsl.Collect (*target, "runtimeenvironment")) ||
	  (target->xrsl.Collect (*target, "middleware")) ||
	  (target->xrsl.Collect (*target, "opsys"))) {
	error = 1;
	goto errorexit;
      }

      if (debug > 1) {
	cout << "Parsed output XRSL:" << endl;
	target->xrsl.Print();
      }

      // dump xrsl

      if (dumpxrsl) {
	cout << target->xrsl.str() << endl;
	submitted = true;
	continue;
      }

      // submit the jobs

      char * shortjobid = NULL;

      int err = ui_uploader (target->cluster->GetContact().c_str(),
			     target->xrsl.str().c_str(), &shortjobid,
			     target->cluster->GetContact().c_str(),
			     RSL_ACTION_REQUEST, filenames, debug);

      if (err || !shortjobid) {
	cout << "Job submission failed to queue " << target->queue->GetName()
	     << " at " << target->cluster->GetLongName() << endl;

	if (shortjobid) {
	  cout << "Deleting partially submitted job" << endl;
	  vector <string> filenames;
	  int err = ui_uploader (target->cluster->GetContact().c_str(), NULL,
				 &shortjobid, NULL, RSL_ACTION_CLEAN,
				 filenames, debug);
	  free (shortjobid);
	}

	cout << "Trying remaining queues" << endl;
	targetlist.erase (target);
	continue;
      }

      if (target->Accept()) {
	error = 1;
	goto errorexit;
      }

      string jobid = target->cluster->GetContact() + '/' + shortjobid;

      free (shortjobid);

      string oldfilename = GetEnv ("HOME");
      oldfilename.append ("/.ngjobs");
      ifstream oldngjobs (oldfilename.c_str());
      string newfilename = oldfilename + ".tmp";
      ofstream newngjobs (newfilename.c_str());
      string jobidname;
      bool written = false;
      while (getline (oldngjobs, jobidname)) {
	int pos = jobidname.find('#');
	string jobname_ = jobidname.substr (pos + 1);
	if (!written && jobname_ > jobname) {
	  newngjobs << jobid << '#' << jobname << endl;
	  written = true;
	}
	newngjobs << jobidname << endl;
      }
      if (!written) newngjobs << jobid << '#' << jobname << endl;
      oldngjobs.close();
      newngjobs.close();
      remove (oldfilename.c_str());
      rename (newfilename.c_str(), oldfilename.c_str());

      string histfilename = GetEnv ("HOME");
      histfilename.append ("/.nghistory");
      ofstream nghist (histfilename.c_str(), ios::app);
      nghist << TimeStamp() << "  " << jobid << endl;
      nghist.close();

      if (!joblist.empty()) {
	ofstream joblist (joblist.c_str(), ios::app);
	joblist << jobid << endl;
	joblist.close();
      }

      cout << "Job submitted with jobid " << jobid << endl;

      submitted = true;
    }
  }

 errorexit:
  DeactivateGlobus();
  return error;
}
